package com.github.polpetta.mezzotre.orm;

import com.google.common.util.concurrent.AbstractExecutionThreadService;
import io.ebean.typequery.PString;
import io.ebean.typequery.TQRootBean;
import io.vavr.Tuple;
import io.vavr.Tuple3;
import io.vavr.control.Try;
import java.util.LinkedList;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.LinkedBlockingDeque;
import java.util.concurrent.TimeUnit;
import java.util.function.Consumer;
import javax.inject.Inject;
import javax.inject.Named;
import org.apache.commons.lang3.tuple.Pair;
import org.slf4j.Logger;

public class BatchBeanCleanerService extends AbstractExecutionThreadService {

  private final LinkedBlockingDeque<
          Tuple3<String, PString<? extends TQRootBean<?, ?>>, CompletableFuture<Integer>>>
      entriesToRemove;
  private final Logger log;
  private final Pair<Integer, TimeUnit> serviceRunningCheckTime;
  private final LinkedList<
          Consumer<Tuple3<String, PString<? extends TQRootBean<?, ?>>, CompletableFuture<Integer>>>>
      deletionListener;

  @Inject
  public BatchBeanCleanerService(
      Logger logger,
      @Named("serviceRunningCheckTime") Pair<Integer, TimeUnit> serviceRunningCheckTime) {
    this.log = logger;
    this.serviceRunningCheckTime = serviceRunningCheckTime;
    this.entriesToRemove = new LinkedBlockingDeque<>();
    this.deletionListener = new LinkedList<>();
  }

  @Override
  protected void run() throws Exception {
    while (isRunning()) {
      // This statement is blocking for 1 sec if the queue is empty, and then it goes on - this way
      // we check if the service is still supposed to be up or what
      Optional.ofNullable(
              entriesToRemove.poll(
                  serviceRunningCheckTime.getLeft(), serviceRunningCheckTime.getRight()))
          .ifPresent(
              entryToRemove -> {
                Try.of(
                        () -> {
                          final int deleted = entryToRemove._2().eq(entryToRemove._1()).delete();
                          if (deleted > 0) {
                            log.trace(
                                "Bean with id " + entryToRemove._1() + " removed successfully ");
                          } else {
                            log.warn(
                                "Bean(s) with id "
                                    + entryToRemove._1()
                                    + " for query "
                                    + entryToRemove._2()
                                    + " was not removed from the database because it was not"
                                    + " found");
                          }
                          entryToRemove._3().complete(deleted);
                          deletionListener.parallelStream().forEach(l -> l.accept(entryToRemove));
                          return null;
                        })
                    .onFailure(ex -> entryToRemove._3().completeExceptionally(ex));
              });
    }
  }

  @Override
  protected void shutDown() throws Exception {
    super.shutDown();
    entriesToRemove.clear();
  }

  public CompletableFuture<Integer> removeAsync(
      String id, PString<? extends TQRootBean<?, ?>> row) {
    final CompletableFuture<Integer> jobExecution = new CompletableFuture<>();
    entriesToRemove.offer(Tuple.of(id, row, jobExecution));
    return jobExecution;
  }

  public void addListener(
      Consumer<Tuple3<String, PString<? extends TQRootBean<?, ?>>, CompletableFuture<Integer>>>
          listener) {
    deletionListener.add(listener);
  }

  public void removeListener(
      Consumer<Tuple3<String, PString<? extends TQRootBean<?, ?>>, CompletableFuture<Integer>>>
          listener) {
    deletionListener.remove(listener);
  }
}
